/* First and next step (assembling module). */

#include "data.h"

#define MINTEMP	10.
#define VGMAX	1.e4

static double delta,rdelta;
static double Ym[EC],Cm[EC];
static double Lvar[EQ],Rvar[EQ],Pl,Pr,Al,Ar,Rhl,Rhr;
static double Cons[EQ];
static double Om[EC],Jac[EC*EC],Dv[EC],Mv[EC*EC];

/* B(t) piston motion */
double Bpiston(double t) {
 double xtvp=vp1*tvp;
 if(t<=tvp) return vp1*t;
 else return xtvp+vp2*(t-tvp);
}

double Vpiston(double t) {
 if(t<=tvp) return vp1;
 else return vp2;
}

/* grid at time = t; delta depends on t. */
static void GetGrid(double t) {
 int k;
 double dx,vt;
 XB[0]=Bpiston(t);XB[nx]=L;
 delta=(XB[nx]-XB[0])/nx;
 for(k=1;k<nx;k++) XB[k]=XB[0]+delta*k;
 vt=Vpiston(t);dx=1./nx;
 for(k=0;k<=nx;k++) UB[k]=vt*(1.-k*dx);
 for(k=1;k<=nx;k++) XC[k]=0.5*(XB[k]+XB[k-1]);
}

static void GetDelta(double t, double dt) {
 delta=(L-Bpiston(t))/nx;
 rdelta=((L-Bpiston(t+dt))/nx)/delta;
}

/* fluxes limiter as a function of 2 variables */
static double FLimiter(double a,double b) {
 return FLimit(a,b,LIMITER,1.5);
}

/* get content of Ua */
static void GetUa(void) {
 int k,c;
 double rho,cp,cv,h0,v;
 for(k=1;k<=nx;k++) {
  for(rho=0.,c=0;c<EC;c++) rho+=Rh(c)[k];
  RH[k]=rho;
  for(c=0;c<EC;c++) Ym[c]=Rh(c)[k]/rho;
  PP[k]=Pressure(TT[k],rho,Ym)+(2./3.)*rho*KA[k];
  GetThermo(&cp,&cv,&h0,Ym);
  AA[k]=e_sqrt(cp/cv*PP[k]/rho);
  VT[k]=Cmu*KA[k]*KA[k]/EP[k];
  v=VV[k];
  h0+=KA[k]+0.5*v*v;
  EE[k]=rho*(cv*TT[k]+h0);
 }
}

/* first step: initial distribution of variables */
void FirstStep(void) {
 int k,c;
 double rh,e0;
 t=0.;dt=0.;
 GetGrid(t);	// get grid
 /* density & mass shares of species */
 Cm[FUEL]=Cfu;Cm[OXID]=Cox;Cm[PROD]=Cpr;Cm[NEUT]=Cnu;
 GetYbyC(Ym,Cm);rh=Density(T0,P0,Ym);
 /* initial epsilon */
 e0=Cmu*Ka0*sqrt(Ka0)/Le0;
 /* fill in the main material arrays */
 for(k=1;k<=nx;k++) {
  TT[k]=T0;
  KA[k]=Ka0;EP[k]=e0;VV[k]=0.;
  PP[k]=P0;
  for(c=0;c<EC;c++) Rh(c)[k]=rh*Ym[c];
 }
 /* other arrays */
 GetUa();
}

/* estimate a new time step. Additional arrs are available */
double TimeStep(void) {
 int k;
 double dtm,v,nu;
 for(dtm=0.,nu=0.,k=1;k<=nx;k++) {
  if((v=fabs(VV[k]+AA[k]))>dtm) dtm=v;
  if((v=fabs(VV[k]-AA[k]))>dtm) dtm=v;
  if((v=(Eta/RH[k]+VT[k]))>nu) nu=v;
 }
 return Sigcour/(dtm/delta+2.*nu/(delta*delta));
}

/* convective & compressive & diffusive fluxes into Fc[] */
static void Fluxes(void) {
 int k,c,k1,km;
 static double ml,mr,m,mpl,ppl,mmr,pmr,a,cp,rh,p,vt,tran,tt,u;
 /* first, obtain flux not mult by velo into Fx */
 for(k=1;k<=nx;k++) {
  for(c=0;c<EC;c++) Fx[c][k]=Rh(c)[k];
  rh=RH[k];
  Fx[VELO][k]=rh*VV[k];
  Fx[KAHA][k]=rh*KA[k];
  Fx[EPSI][k]=rh*EP[k];
  Fx[TEMP][k]=EE[k]+PP[k];
 }
 /* cycle internal cells interfaces */
 for(k=1;k<nx;k++) {
  k1=k+1;
  /* sonic velocity */
  a=e_sqrt(AA[k]*AA[k1]);
  /* relative Mach numbers */
  ml=(VV[k]-UB[k])/a;mr=(VV[k1]-UB[k])/a;
  /* get Mach number & pressure cofs split */
  AusmPlus(ml,&mpl,&ppl);AusmMinus(mr,&mmr,&pmr);
  /* interface Mach, interface velo */
  m=(mpl+mmr)*a;
  /* upwind scheme by Liou: convective term */
  if(m>=0.) {for(c=0;c<EQ;c++) Fc[c][k]=Fx[c][k]*m;}
  else {for(c=0;c<EQ;c++) Fc[c][k]=Fx[c][k1]*m;}
  /* add compressive term & add to pressure work */
  p=ppl*PP[k]+pmr*PP[k1];
  Fc[VELO][k]+=p;
  Fc[TEMP][k]+=p*UB[k];
 }
 /* the leftmost interface */
 k=0;k1=1;
 /* zero convective flux */
 for(c=0;c<EQ;c++) Fc[c][k]=0.;
 /* compressive flux */
 Fc[VELO][k]+=PP[k1];
 Fc[TEMP][k]+=PP[k1]*UB[k];
 /* the rightmost flux */
 k=nx;
 /* zero convective flux */
 for(c=0;c<EQ;c++) Fc[c][k]=0.;
 /* compressive flux */
 Fc[VELO][k]+=PP[k];
 Fc[TEMP][k]+=PP[k]*UB[k];
 /* upgrade Fx (added U*0.5(UB[k]+UB[k-1])) */
 for(k=1;k<=nx;k++) {
  vt=VV[k];
  for(c=0;c<EQ;c++) Fx[c][k]*=vt;
  Fx[VELO][k]+=PP[k];
 }
 /* upgrade flux */
 for(c=0;c<EQ;c++) {
  double *fx=Fx[c],*fc=Fc[c];
  if(c==KAHA||c==EPSI) continue;
  for(k=0;k<=nx;k++) {
   if((k1=k+1)>nx) k1=nx;
   if((km=k)<1) km=1;
   Ftmp[k]=fc[k];
   m=0.5*(VV[k1]+VV[km])-UB[k];
   if(m>=0.) {
    switch(c) {
     case VELO: Ftmp[k]+=UB[k]*RH[k1]*VV[k1];break;
     case KAHA: Ftmp[k]+=UB[k]*RH[k1]*KA[k1];break;
     case EPSI: Ftmp[k]+=UB[k]*RH[k1]*EP[k1];break;
     case TEMP: Ftmp[k]+=UB[k]*EE[k1];break;
     default:   Ftmp[k]+=UB[k]*Rh(c)[k1];break;
    }
   }
   else {
    switch(c) {
     case VELO: Ftmp[k]+=UB[k]*RH[km]*VV[km];break;
     case KAHA: Ftmp[k]+=UB[k]*RH[km]*KA[km];break;
     case EPSI: Ftmp[k]+=UB[k]*RH[km]*EP[km];break;
     case TEMP: Ftmp[k]+=UB[k]*EE[km];break;
     default:   Ftmp[k]+=UB[k]*Rh(c)[km];break;
    }
   }
  }
  for(k=1;k<nx;k++) fc[k]+=0.5*(FLimiter(fx[k]-Ftmp[k-1],fx[k+1]-Ftmp[k])
	+FLimiter(fx[k+1]-Ftmp[k+1],fx[k]-Ftmp[k]));
 }
 /* subtract diffusive terms from Fc */
 for(k=1;k<nx;k++) {
  k1=k+1;
  vt=0.5*(RH[k]*VT[k]+RH[k1]*VT[k1]);
  tran=(vt/SIGT+Eta)/delta;
  for(c=0;c<EC;c++) Fc[c][k]-=tran*(Rh(c)[k+1]/RH[k+1]-Rh(c)[k]/RH[k]);
  tran=(vt+Eta)*(4./3.)/delta;
  Fc[VELO][k]-=tran*(VV[k+1]-VV[k]);
  tran=(vt+Eta)/delta;
  Fc[KAHA][k]-=tran*(KA[k+1]-KA[k]);
  tran=(vt/Sige+Eta)/delta;
  Fc[EPSI][k]-=tran*(EP[k+1]-EP[k]);
  for(c=0;c<EC;c++) Ym[c]=(Rh(c)[k+1]+Rh(c)[k])/(RH[k]+RH[k+1]);
  GetThermo(&cp,0,0,Ym);
  tran=cp*(vt/PRANT+Eta/Pran)/delta;
  Fc[TEMP][k]-=tran*(TT[k+1]-TT[k]);
  tt=0.5*(TT[k]+TT[k+1]);
  tran=(vt/SIGT+Eta/Schm)/delta;
  for(c=0;c<EC;c++) Fc[TEMP][k]-=tran*(Cpm[c]*tt+Hm[c])*
	(Rh(c)[k+1]/RH[k+1]-Rh(c)[k]/RH[k]);
  tran=(vt+Eta)/delta;
  Fc[TEMP][k]-=tran*(KA[k+1]-KA[k]);
 }
} 

static void GetReaction(double om[], double jac[], double rh[], double fr) {
 int k,j,i;
 double omg,jg;
 omg=rh[FUEL]*rh[OXID]*fr;
 for(k=0;k<EC;k++) {
  om[k]=Csw[k]*omg;jg=Csw[k]*fr;
  for(j=0,i=k*EC;j<EC;j++,i++) jac[i]=0.;
  i=k*EC;
  jac[i+FUEL]=jg*rh[OXID];
  jac[i+OXID]=jg*rh[FUEL];
 }
}

static double Chemist(double s[], double T0, double cons[], double Ehat) {
 int c,j,i,it=0;
 static double zdt,fr,om,tmp,lam,cv;
 zdt=0.5*(1.+1/rdelta)*dt;
 fr=Funarr1(T0);
 do {
  GetReaction(Om,Jac,s,fr);
  for(c=0;c<EC;c++) {
   Dv[c]=cons[c]-s[c]+zdt*Om[c];
   for(j=0,i=c*EC;j<EC;j++,i++) {
    Mv[i]=-zdt*Jac[i];
    if(j==c) Mv[i]+=1.;
   }
  }
  /* solve */
  LinSolve(EC,Dv,Mv);
  for(lam=1.,c=0;c<EC;c++) if((tmp=2.*fabs(Dv[c])/(cons[c]+ATOL))>lam) lam=tmp;
  for(c=0;c<EC;c++) s[c]+=Dv[c]/lam;
  for(c=0;c<EC;c++) if(s[c]<0.) s[c]=0.;
  ++it;
 } while(islongstep(EC,Dv,s) && it<500);
 for(c=0;c<EC;c++) Ehat-=s[c]*(Cvm[c]*T0+Hm[c]);
 return Ehat;
}

static double GetTemp(double s[], double Ehat) {
 int c;
 double cv=0.;
 for(c=0;c<EC;c++) {Ehat-=s[c]*Hm[c];cv+=s[c]*Cvm[c];}
 return Ehat/cv;
}

static void ChemStep(double *T, double rh[], double cons[], double Ehat) {
 int k;
 static double fu0,te0,fu1,te1,dT,fu,dir,fL,fR,tL,tR;
 te0=(*T);
 fu0=Chemist(rh,te0,cons,Ehat);
 te1=GetTemp(rh,Ehat);
 if(fabs(fu0)<RTOL*fabs(Ehat)+ATOL) {*T=te1;return;}
// *T=te1;return;
 /* secant method stage */
 do {
  fu1=Chemist(rh,te1,cons,Ehat);
  if(fabs(fu1)<RTOL*fabs(Ehat)+ATOL) {*T=te1;return;}
  if(fabs(te1-te0)<RTOL*fabs(te1)+ATOL) {*T=te1;return;}
  if(fabs(fu1-fu0)<RTOL*fabs(fu0)+ATOL) break;
  if(fabs(fu1)>fabs(fu0)*DIMRATE) break;
  dT=-fu1*(te1-te0)/(fu1-fu0);
  te0=te1;te1=te0+dT;fu0=fu1;
 } while(1);
 /* root bracketing stage */
 fL=fu0;tL=te0;fR=fu1;tR=te1;
 if(fL*fR<0.) goto DICHO;	// to the dichotomy stage
 dir=fabs(tR-tL);
 if(fL<0.) dir=-dir;
 fL=fR=fu1;tL=tR=te1;
 do {
  tR+=dir;
  fR=Chemist(rh,tR,cons,Ehat);
  if(fabs(fR)<RTOL*fabs(Ehat)+ATOL) {*T=tR;return;}
  if(fL*fR<0) break;
  fL=fR;tL=tR;dir*=BRKRATE;
 } while(1);
 /* dichotomy stage */
DICHO:
 while(fabs(tR-tL)>RTOL*(*T)+ATOL) {
  *T=0.5*(tL+tR);
  fu=Chemist(rh,*T,cons,Ehat);
  if(fu*fL>0.) {tL=(*T);fL=fu;}
  else if(fu*fR>0.) {tR=(*T);fR=fu;}
  else return;
 }
 *T=0.5*(tL+tR);
}

/* promote variables */
void Promote_vars(int k) {
 int c,km,kp;
 double tmp;
 static double rh,cp,cv,h0,vg,pi1,pi2,ga1,ga2,ehat,s[EC],lam;
 /* get conservs */
 rh=RH[k];
 for(c=0;c<EC;c++) Cons[c]=Rh(c)[k];
 Cons[VELO]=rh*VV[k];
 Cons[KAHA]=rh*KA[k];
 Cons[EPSI]=rh*EP[k];
 Cons[TEMP]=EE[k];
 /* former velocity gradient (central) */
 if((km=k-1)<1) km=1;
 if((kp=k+1)>nx) kp=nx;
 vg=(VV[kp]-VV[km])/(XC[kp]-XC[km]);
 if(vg>VGMAX) vg=VGMAX;
 if(vg<-VGMAX) vg=-VGMAX;
 /* apply fluxes */
 for(c=0;c<EQ;c++) Cons[c]-=(Fc[c][k]-Fc[c][k-1])*dt/delta;
 /* add the external heat flux */
 if(is_Tp==0 && k==1) Cons[TEMP]+=qw*dt/delta;
 /* apply external fluxes & resolve */
 for(c=0;c<EQ;c++) Cons[c]/=rdelta;
 /* density */
 for(c=0,rh=0.;c<EC;c++) rh+=Cons[c];
 /* velocity */
 VV[k]=Cons[VELO]/rh;
 /* boundary conditions, velocity */
 if(k==1) VV[k]=(VV[k+1]*0.5+UB[k-1])/1.5;
 else if(k==nx) VV[k]=(VV[k-1]*0.5+UB[k])/1.5;
 /* promote turbulence */
 Cons[KAHA]/=rh;Cons[EPSI]/=rh;
 if(Cons[KAHA]<0.5*KA[k]) Cons[KAHA]=0.5*KA[k];
 if(Cons[KAHA]>2.0*KA[k]) Cons[KAHA]=2.0*KA[k];
 if(Cons[EPSI]<0.5*EP[k]) Cons[EPSI]=0.5*EP[k];
 if(Cons[EPSI]>2.0*EP[k]) Cons[EPSI]=2.0*EP[k];
//goto NOTUR;
 pi1=4.*Cmu*vg*vg/3.;ga2=0.5*(1.+1/rdelta);
 s[0]=KA[k];s[1]=EP[k];
 do {
  Dv[0]=Cons[KAHA]/dt-s[0]/dt+ga2*(pi1*s[0]*s[0]/s[1]-s[1]);
  Dv[1]=Cons[EPSI]/dt-s[1]/dt+ga2*(C1e*pi1*s[0]-C2e*s[1]*s[1]/s[0]);
  Mv[0]=1./dt-ga2*(2.*pi1*s[0]/s[1]);
  Mv[1]=ga2*(pi1*(s[0]*s[0])/(s[1]*s[1])+1.);
  Mv[2]=-ga2*(C1e*pi1+C2e*(s[1]*s[1])/(s[0]*s[0]));
  Mv[3]=1./dt+ga2*(2.*C2e*s[1]/s[0]);
  LinSolve2(Dv,Mv);
  lam=1.;
  if((tmp=2.*fabs(Dv[0])/s[0])>lam) lam=tmp;
  if((tmp=2.*fabs(Dv[1])/s[1])>lam) lam=tmp;
  s[0]+=Dv[0]/lam;s[1]+=Dv[1]/lam;
 } while(islongstep(2,Dv,s));
 KA[k]=(s[0]>KAMIN)?s[0]:KAMIN;EP[k]=(s[1]>EPMIN)?s[1]:EPMIN;
//NOTUR: KA[k]=Cons[KAHA];EP[k]=Cons[EPSI];
 /* temperature & concentrations */
 Cons[TEMP]-=rh*(0.5*VV[k]*VV[k]+KA[k]);
 /* case1: zero chemistry */
 if(TT[k]<Tm) {
  for(c=0;c<EC;c++) Rh(c)[k]=Cons[c];
  TT[k]=GetTemp(Cons,Cons[TEMP]);
 }
 /* chemistry switched on */
 else {
  for(c=0;c<EC;c++) s[c]=Rh(c)[k];
  ChemStep(TT+k,s,Cons,Cons[TEMP]);
  for(c=0;c<EC;c++) Rh(c)[k]=s[c];
 }
 /* boundary conditions & troubleshooting */
 if(k==1) {
  if(is_Tp) TT[k]=(TT[k+1]*0.5+Tp)/1.5;
  else if(qw>=0. && TT[k]<TT[k+1]) TT[k]=TT[k+1];
  else if(qw<=0. && TT[k]>TT[k+1]) TT[k]=TT[k+1];
  if(TT[k]<MINTEMP) TT[k]=MINTEMP;
 }
}

/* one step */
void NextStep(void) {
 int k;
 dt=TimeStep();GetDelta(t,dt);
 /* calculate fluxes */
 Fluxes();
 /* promote variables */
 for(k=2;k<nx;k++) Promote_vars(k);
 Promote_vars(1);Promote_vars(nx);
 GetUa();
 /* promote time */
 t+=dt;
 GetGrid(t);
}


