/* utmath.c - low-level math utilities */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "utmath.h"

double e_sqrt(double x) {
 if(x<=0.) return 0.;
 else return sqrt(x);
}

/* condition is true when relative dv[k] is too high */
int islongstep(int m,double *dv,double *v) {
 int q;
 for(q=0;q<m;q++) if(ATOL+RTOL*fabs(v[q])<fabs(dv[q])) return 1;
 return 0;
}

/* linear system of 2 equations (Kramer rule) */
int LinSolve2(double *dv,double *mv) {
 double d=mv[0]*mv[3]-mv[1]*mv[2],tv0=dv[0];
 if(d==0.) return 0.;
 dv[0]=(tv0*mv[3]-dv[1]*mv[1])/d;
 dv[1]=(dv[1]*mv[0]-tv0*mv[2])/d;
 return 1;
}

/* Gaussian method */
int LinSolve(int n,double *dv,double *mv) {
 int i,j,k,in,kn;
 double t,p;
 if(n<=0) return 0;
 /* first pass of Gaussian elimination */
 for(i=0;i<n-1;i++) {
  /* seek for row with max |mv[i][k>=i]| */
  k=i;p=fabs(mv[i*n+i]);
  for(j=i+1;j<n;j++) if((t=fabs(mv[j*n+i]))>p) {p=t;k=j;}
  /* exchange rows i and k */
  in=i*n;
  if(k>i) {
   kn=k*n;
   for(j=i;j<n;j++) {t=mv[kn+j];mv[kn+j]=mv[in+j];mv[in+j]=t;}
   t=dv[k];dv[k]=dv[i];dv[i]=t;
  }
  p=mv[in+i];
  if(p==0.) return 0;	/* singularity */
  /* eliminate from the rows i+1,...,n-1 */
  for(k=i+1;k<n;k++) {
   kn=k*n;
   t=mv[kn+i]/p;mv[kn+i]=0.;
   for(j=i+1;j<n;j++) mv[kn+j]-=t*mv[in+j];
   dv[k]-=t*dv[i];
  }
 }
 /* second pass */
 for(i=n-1;i>=0;i--) {
  in=i*n;
  for(j=i+1;j<n;j++) dv[i]-=mv[in+j]*dv[j];
  dv[i]/=mv[in+i];
 }
 return 1;
}

/* minmod function */
double Minmod(double a,double b) {
 if((a<=0.&&b>=0.)||(a>=0.&&b<=0.)) return 0.;
 else if(fabs(a)<fabs(b)) return a;
 else return b;
}

double VanLeer(double a,double b) {
 if((a<=0.&&b>=0.)||(a>=0.&&b<=0.)) return 0.;
 else return 2.*a*b/(a+b);
}

double Superbee(double a,double b) {
 double fa=fabs(a),fb=fabs(b);
 if((a>=0.&&b<=0.)||(a<=0.&&b>=0.)) return 0.;
 else if(fa>2.*fb) return 2.*b;
 else if(fa>fb) return a;
 else if(fa>0.5*fb) return b;
 else return 2.*a;
}

/* general flux limiter function */
double FLimit(double a,double b,int name,double addpar) {
 double r,t;
 /* common limiters */
 switch(name) {
  case 0: return 0.;
  case MINMOD: return Minmod(a,b);
  case VANLEER: return VanLeer(a,b);
  case SUPERBEE: return Superbee(a,b);
 }
 /* other limiters */
 if((a>=0.&&b<=0.)||(a<=0.&&b>=0.)) return 0.;
 r=a/b;
 switch(name) {
  case KOREN:
   a=2.;
   if((t=(1.+2.*r)/3.)<a) a=t;
   if((t=2.*r)<a) a=t;
   return a*b;
  case MCLIM:
   a=2.;
   if((t=0.5*(1.+r))<a) a=t;
   if((t=2.*r)<a) a=t;
   return a*b;
  case OSHER:
   a=addpar;
   if((t=r)<a) a=t;
   return a*b;
  case OSPRE:
   a=1.5*r*(r+1.)/(r*(r+1.)+1);
   return a*b;
  case UMIST:
   a=2.;
   if((t=0.75+0.25*r)<a) a=t;
   if((t=0.25+0.75*r)<a) a=t;
   if((t=2.*r)<a) a=t;
   return a*b;
  case VANALBADA:
   a=r*(r+1.)/(r*r+1.);
   return a*b;
  case SWEBY:
   a=addpar;t=addpar*r;
   if(r<a) a=r;
   if(1<t) t=1.;
   if(t>a) a=t;
   return a*b;
  default: return 0.;
 }
}

/* Ausm split cofs. */
void AusmPlus(double m,double *map,double *pip) {
 if(m>=1.) {*map=m;*pip=1.;}
 else if(m<=-1.) {*map=0.;*pip=0.;}
 else {
  double m1=m+1.,m2=m*m-1.;
  *map=0.25*m1*m1+0.125*m2*m2;
  *pip=0.25*m1*m1*(2.-m)+(3./16.)*m*m2*m2;
 }
}

void AusmMinus(double m,double *mam,double *pim) {
 if(m>=1.) {*mam=0.;*pim=0.;}
 else if(m<=-1.) {*mam=m;*pim=1.;}
 else {
  double m1=m-1.,m2=m*m-1.;
  *mam=-0.25*m1*m1-0.125*m2*m2;
  *pim=0.25*m1*m1*(2.+m)-(3./16.)*m*m2*m2;
 }
}

/* coefficient of the variables change */
double Change(double *dv,double *va,int n) {
 int k;
 double lam,var,dvar,tmp;
 lam=0.;
 for(k=0;k<n;k++) {
  var=fabs(va[k])+ATOL;dvar=fabs(dv[k]);
  if(lam<(tmp=dvar/var)) lam=tmp;
 }
 return lam;
}



